#include "llc_iface.h"
#include "firewire.h"
#include "firecam.h"

static int cameras[2];
int firecam_num_cameras = 0;

// Registers
#define CAM_REG(x) (0xFFFFF0F00000LL | (x))

#define CAM_INITIALIZE		0x000
#define V_FORMAT_INQ		0x100
#define V_MODE_INQ_BASE		0x180
#define V_RATE_INQ_BASE		0x200
#define FEATURE_HI_INQ		0x404
#define FEATURE_LO_INQ		0x408
#define ELEMS_INQ_BASE		0x500
#define CUR_V_FRM_RATE		0x600
#define CUR_V_MODE		0x604
#define CUR_V_FORMAT		0x608
#define ISO_MODE		0x60C
#define CAMERA_POWER		0x610
#define ISO_ENABLE		0x614
#define ELEMS_CTRL_BASE		0x800

static void write_quad(int node, long long int addr, unsigned int data)
{
	// Create a packet
	llc_packet* pkt = (llc_packet*) malloc(sizeof(llc_packet));
	
	// Allocate 1-quadlet of data and configure this
	// packet ot be a "quadlet write-request" packet type
	pkt->data = (int*) malloc(1 * sizeof(int));
	pkt->spd = SPD_400;
	pkt->tlabel = 0;
	pkt->rt = 0;
	pkt->tcode = TC_REQ_QUAD_WRITE;
	pkt->dest_id = node;
	pkt->data_length = 1;
	pkt->dest_offset = addr;
	*pkt->data = data;
	
	llc_send_packet(pkt);
	llc_free_packet(pkt);

	if (node == NODE_BROADCAST)
	{
		while (llc_packet_available())
			llc_free_packet(llc_recv_packet());

		return;
	}
	else if (llc_get_last_ack() == ACK_PENDING)
	{
		while (!llc_packet_available())
			;

		llc_free_packet(llc_recv_packet());

		return;
	}
	else
	{
		return;
	}
}



static unsigned int read_quad(int node, long long int addr)
{
	int data;

	// Create a request packet
	llc_packet* pkt = (llc_packet*) malloc(sizeof(llc_packet));
	
	// Configure this
	// packet ot be a "quadlet read-request" packet type
	pkt->spd = SPD_100;
	pkt->tlabel = 0;
	pkt->rt = 0;
	pkt->tcode = TC_REQ_QUAD_READ;
	pkt->dest_id = node;
	pkt->data_length = 0;
	pkt->dest_offset = addr;
	
	llc_send_packet(pkt);
	llc_free_packet(pkt);

	// Wait for response
	while (!llc_packet_available())
		;

	pkt = llc_recv_packet();
	data = *pkt->data;
	llc_free_packet(pkt);

	return data;
}



static void reset_llc_iface()
{
	// Data-Mover configuration:
	// Configure datamover for recieve-only
	// of isochronous packets. 
	llc_write(0x04, 0x0000);
	llc_write(0x06, 0x0225);

	// General configuration:
	// Reset TX/RX, enable
	// cyclemaster, enable isochronous
	// receive for two channels.
	llc_write(0x08, 0xc6f5);
	llc_write(0x0a, 0x0B0E);

	// Interrupt register/mask register:
	// Clear all interrupts and
	// receive all interrupts
	llc_write(0x0c, 0xffff);
	llc_write(0x0e, 0xffff);
	llc_write(0x10, 0xffff);
	llc_write(0x12, 0xffff);

	// Isochronous configuration:
	// Use channel #0 and #1
	// Monitor packets for "sync"
	// value of 0001b which the
	// camera puts out when it's
	// the start of a new frame.
	llc_write(0x18, 0x0001);
	llc_write(0x1a, 0x0010);
	llc_write(0x30, 0x8004);
	llc_write(0x24, 0x44c1);
	llc_write(0x24, 0x4503);

	// Set up Retry Register
	llc_write(0x4c, 0xff01);
}



//
// Public Functions
//

int firecam_open()
{
	llc_init();
	firecam_num_cameras = 2;
	cameras[0] = 0xFFC0;
	cameras[1] = 0xFFC1;

	return 0;
}

int firecam_init()
{
	reset_llc_iface();

	return 0;
}


int firecam_enumerate()
{
	int i;
	llc_selfid_packet* selfpkts;
	
	// PHY Access Register:
	// Reset the bus and become root node
	llc_write(0x24, 0x41ff);
	// Re-enables transmitter after bus reset
	llc_write(0x08, 0xc6c5);

	// Wait for self-id packets to come in
	while (!llc_packet_available())
		;
	
	// Get self-id packets after bus reset
	selfpkts = llc_recv_selfid_packets();

	firecam_num_cameras = selfpkts[0].num_pkts - 1;
	
	for (i = 0; i < selfpkts[0].num_pkts; i++)
	{
		if (PHY_ID(llc_get_node_id()) != selfpkts[i].phy_id)	
			cameras[i] = (0xFFC0 | selfpkts[i].phy_id);
	}

	free(selfpkts);

	return 0;
}


int firecam_reset(int cam_id)
{
	write_quad(cameras[cam_id], CAM_REG(CAM_INITIALIZE), 0x80000000);

	return 0;
}


int firecam_format_available(int cam_id, int format)
{
	unsigned int data;

	data = read_quad(cameras[cam_id], CAM_REG(V_FORMAT_INQ));

	return (data & (0x80000000 >> format)) != 0;
}


int firecam_mode_available(int cam_id, int format, int mode)
{
	unsigned int data;
	long long int addr;

	addr = CAM_REG(V_MODE_INQ_BASE) + 4*format;
	data = read_quad(cameras[cam_id], addr);

	return (data & (0x80000000 >> mode)) != 0;
}


int firecam_rate_available(int cam_id, int format, int mode, int rate)
{
	unsigned int data;
	long long int addr;

	addr = CAM_REG(V_RATE_INQ_BASE) + 0x20*mode + 4*rate;
	data = read_quad(cameras[cam_id], addr);
	
	return (data & (0x80000000 >> rate)) != 0;
}


int firecam_set_format(int cam_id, int format)
{
	write_quad(cameras[cam_id], CAM_REG(CUR_V_FORMAT), format << 29);

	return 0;
}


int firecam_set_mode(int cam_id, int mode)
{
	write_quad(cameras[cam_id], CAM_REG(CUR_V_MODE), mode << 29);

	return 0;
}


int firecam_set_rate(int cam_id, int rate)
{
	write_quad(cameras[cam_id], CAM_REG(CUR_V_FRM_RATE), rate << 29);

	return 0;
}


int firecam_start_streaming(int cam_id, int speed)
{
	int data;

	data = (cam_id<<28) | (speed<<24);
	
	write_quad(cameras[cam_id], CAM_REG(ISO_MODE), data);
	write_quad(cameras[cam_id], CAM_REG(ISO_ENABLE), 0x80000000);

	return 0;
}


int firecam_stop_streaming(int cam_id)
{
	write_quad(cameras[cam_id], CAM_REG(ISO_ENABLE), 0x0);

	return 0;
}


int firecam_is_streaming(int cam_id)
{
	return read_quad(cameras[cam_id], CAM_REG(ISO_ENABLE)) != 0;
}


int firecam_query_element(int cam_id, int element_id, elem_query_t* out)
{
	long long int addr;
	unsigned int data;

	addr = CAM_REG(ELEMS_INQ_BASE) +  4*element_id;
	data = read_quad(cameras[cam_id], addr);

	out->available = (data & 0x80000000) != 0;
	out->auto_control = (data & 0x02000000) != 0;
	out->man_control = (data & 0x01000000) != 0;
	out->on_off = (data & 0x04000000) != 0;
	out->min_value = (data >> 12) & 0xFFF;
	out->max_value = data & 0xFFF;

	return 0;
}


int firecam_get_elem_status(int cam_id, int element_id, elem_control_t* out)
{
	long long int addr;
	unsigned int data;

	addr = CAM_REG(ELEMS_CTRL_BASE) + 4*element_id;
	data = read_quad(cameras[cam_id], addr);

	out->on_off = (data & 0x02000000) != 0;
	out->auto_control = (data & 0x01000000) != 0;
	out->value = data & 0xFFF;
	out->value2 = (data >> 12) & 0xFFF;
	
	return 0;
}


int firecam_set_elem_status(int cam_id, int element_id, elem_control_t* in)
{
	long long int addr;
	unsigned int data;

	addr = CAM_REG(ELEMS_CTRL_BASE) + 4*element_id;
	data = 0;
	data |= in->on_off << 26;
	data |= in->auto_control << 25;
	data |= in->value;

	if (element_id == ELEM_WHITE_BALANCE)
		data |= in->value2 << 12;

	write_quad(cameras[cam_id], addr, data);

	return 0;
}
