--
-- firevid_decoder_mono8.vhd
--
-- This module takes raw 32-bit data quadlets
-- from a Firewire digital camera and generates
-- RGB/row/column signals from the stream.
--
-- This module decodes an 8-bit monochrome stream
--
-- ---------------------------------------------------------
--
-- In MONO8 format, every byte is an unsigned brightness value
-- from 0-255. The data stream cmoing from the camera looks
-- like this:
--
-- Quadlet 0 : Y0 Y1 Y2 Y3
-- Quadlet 1 : Y4 Y5 Y6 Y7
-- ...
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity firevid_decoder_mono8 is
generic
(
	WIDTH		: natural := 640;
	HEIGHT		: natural := 480
);
port
(
	clock		: in	std_logic;
 	reset_n		: in	std_logic;

	-- Signals from DataMover-FIFO module
	read_ack	: out	std_logic;
	quadlet_valid	: in	std_logic;
	quadlet_data	: in	std_logic_vector(31 downto 0);
	new_frame	: in	std_logic;

	-- User signals
	row		: buffer	std_logic_vector(9 downto 0);
	column		: buffer	std_logic_vector(9 downto 0);
	red		: out	std_logic_vector(7 downto 0);
	green		: out	std_logic_vector(7 downto 0);
	blue		: out	std_logic_vector(7 downto 0);
	data_valid	: out	std_logic
);
end firevid_decoder_mono8;


architecture rtl of firevid_decoder_mono8 is
	signal new_frame_r	: std_logic;
begin
	-- We acknowledge the current quadlet (and thus ask to
	-- retrieve a new one for next cycle) during the 4th
	-- pixel of every 4-pixel group.
	read_ack <= '1' when (column(1 downto 0) = "11") else '0';

	-- As long as we have a quadlet to process, we are outputting
	-- a valid pixel.
	data_valid <= quadlet_valid;

	-- This mux selects which byte to use from the incoming
	-- data quadlet depending on which pixel (0-3) we are
	-- processing.
	process (column, quadlet_data)
	begin
		case column(1 downto 0) is
		when "00" =>
			red <= quadlet_data(31 downto 24);
			green <= quadlet_data(31 downto 24);
			blue <= quadlet_data(31 downto 24);
		when "01" =>
			red <= quadlet_data(23 downto 16);
			green <= quadlet_data(23 downto 16);
			blue <= quadlet_data(23 downto 16);
		when "10" =>
			red <= quadlet_data(15 downto 8);
			green <= quadlet_data(15 downto 8);
			blue <= quadlet_data(15 downto 8);
		when "11" =>
			red <= quadlet_data(7 downto 0);
			green <= quadlet_data(7 downto 0);
			blue <= quadlet_data(7 downto 0);
		end case;
	end process;

	-- Save current value of new_frame so we can edge detect
	process (reset_n, clock, new_frame)
	begin
		if reset_n = '0' then
			new_frame_r <= '0';
		elsif falling_edge(clock) then
			new_frame_r <= new_frame;
		end if;
	end process;

	process (reset_n, clock, new_frame)
	begin
		if reset_n = '0' or (new_frame = '1' and new_frame_r = '0') then
			column <= (others => '0');
			row <= (others => '0');
		elsif rising_edge(clock) then
			-- Advance the output row/column counters.
			if quadlet_valid = '1' then
				if column = (WIDTH-1) then
					row <= row + 1;
					column <= (others => '0');
				else
					column <= column + 1;
				end if;
			end if;
		end if;
	end process;
end rtl;
